"""
tictail.resource.definitions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Definitions for all API endpoints and their corresponding instances.

"""

from .base import (Collection,
                   Resource,
                   Get,
                   GetById,
                   List,
                   Create,
                   Delete,
                   DeleteById)


class Follower(Resource, Delete):
    endpoint = 'followers'


class Followers(Collection, List, Create, DeleteById):
    resource = Follower


class Product(Resource, Get):
    endpoint = 'products'


class Products(Collection, GetById, List):
    resource = Product

    def format_params(self, **params):
        if 'categories' in params:
            params['categories'] = ','.join(params['categories'])
        return params


class Card(Resource):
    endpoint = 'cards'


class Cards(Collection, Create):
    resource = Card


class Customer(Resource, Get):
    endpoint = 'customers'


class Customers(Collection, GetById, List):
    resource = Customer


class Order(Resource, GetById):
    endpoint = 'orders'


class Orders(Collection, GetById, List):
    resource = Order

    def _isoformat(self, dt):
        try:
            return dt.isoformat()
        except AttributeError:
            return dt

    def format_params(self, **params):
        modified_before = params.get('modified_before')
        if modified_before:
            params['modified_before'] = self._isoformat(modified_before)

        modified_after = params.get('modified_after')
        if modified_after:
            params['modified_after'] = self._isoformat(modified_after)

        return params


class Theme(Resource, Get):
    endpoint = 'theme'
    singleton = True


class Category(Resource):
    endpoint = 'categories'


class Categories(Collection, List):
    resource = Category


class Store(Resource, Get):
    endpoint = 'stores'
    subresources = [
        Cards,
        Products,
        Customers,
        Followers,
        Orders,
        Theme,
        Categories
    ]


class Stores(Collection, GetById):
    resource = Store


class Me(Store, Get):
    endpoint = 'me'
    singleton = True

    def instantiate_from_data(self, data):
        return Store(self.transport, data=data)


__all__ = [
    'Follower', 'Followers', 'Product', 'Products', 'Card', 'Cards',
    'Customer', 'Customers', 'Order', 'Orders', 'Theme', 'Category',
    'Categories', 'Store', 'Stores', 'Me'
]



"""
tictail.transport
~~~~~~~~~~~~~~~~~

Strategies for communication with the API. For now, communication is only
done via HTTP via the `requests` library. For documentation on `requests`,
please visit: http://docs.python-requests.org/en/latest/.

"""

from .version import __version__
from .importer import json, requests
from .errors import (ApiConnectionError,
                     ApiError,
                     Forbidden,
                     NotFound,
                     BadRequest,
                     ServerError)


ConnectionError = requests.exceptions.ConnectionError
HTTPError = requests.exceptions.HTTPError


class RequestsHttpTransport(object):
    """Handles communication and data mungling.

    Internally, it uses `requests` to issue http requests to the various
    endpoints, and only speaks JSON.

    """

    def __init__(self, access_token, config):
        self.access_token = access_token
        self.config = config

    def _make_abs_uri(self, uri):
        """Makes an absolute API uri using the API base and protocol.

        :param uri: The URI to absolutize.

        """
        base = self.config['base']
        version = "v{0}".format(self.config['version'])
        protocol = self.config['protocol']
        if uri[0] == '/':
            uri = uri[1:]
        if uri[-1] == '/':
            uri = uri[:-1]
        return "{0}://{1}/{2}/{3}".format(protocol, base, version, uri)

    def _utf8(self, value):
        return value.encode('utf-8') if isinstance(value, unicode) else value

    def _handle_connection_error(self, err):
        raise ApiConnectionError(err.message)

    def _handle_http_error(self, err):
        resp = err.response
        status_code = resp.status_code

        try:
            resp_json = resp.json()
            message = resp_json.get('message')
        except ValueError:
            # If we can't read JSON from the error response, then something is
            # obviously wrong. Check for Bad Gateway errors first, otherwise
            # just show a generic error message and raise a `ServerError`.
            if status_code == 502:
                message = ('It seems that the API is unreachable. Please contact '
                           'Tictail Support if the problem persists.')
            else:
                message = ('An unexpected error occured. Please contact Tictail '
                           'Support if the problem persists.')

            raise ServerError(message, status_code, resp.text)

        # Raise appropriate exception for 400, 403, 404, and a generic error
        # for all other error codes.
        if status_code == 400:
            err_cls = BadRequest
        elif status_code == 403:
            err_cls = Forbidden
        elif status_code == 404:
            err_cls = NotFound
        else:
            err_cls = ServerError

        raise err_cls(message, status_code, resp.text, json=resp_json)

    def _handle_unexpected_error(self, err):
        if isinstance(err, ValueError):
            resp = err.response
            # For a `ValueError` to occur, it might mean that the API did not
            # return valid JSON as expected, even though no HTTP error occured.
            if 'JSON object' in str(err):
                content_type = resp.headers['content-type']
                message = ("The API should return JSON, but there was a problem "
                           "decoding it. The response content-type was: `{0}`."
                           .format(content_type))
                raise ApiError(message, resp.status_code, resp.text)

        # Just raise the exception if we can't handle it in a better way.
        raise err

    def get(self, uri, params=None):
        """Issues a GET request.

        :param uri: The resource URI.
        :param params: Query parameters as dict or bytes.

        """
        return self.handle_request('GET', uri, params=params)

    def post(self, uri, params=None, data=None):
        """Issues a POST request.

        :param uri: The resource URI.
        :param params: Query parameters as dict or bytes.
        :param data: A dictionary representing a JSON body.

        """
        return self.handle_request('POST', uri, params=params, data=data)

    def put(self, uri, params=None, data=None):
        """Issues a PUT request.

        :param uri: The resource URI.
        :param params: Query parameters as dict or bytes.
        :param data: A dictionary representing a JSON body.

        """
        return self.handle_request('PUT', uri, params=params, data=data)

    def delete(self, uri, params=None):
        """Issues a DELETE request.

        :param uri: The resource URI.
        :param params: Query parameters as dict or bytes.

        """
        return self.handle_request('DELETE', uri, params=params)

    def handle_request(self, method, uri, params=None, data=None):
        """The bread and butter of this class. Issues a HTTP requests and takes
        care of all errors. Returns the JSON-decoded data and the status code.

        :param method: The HTTP method to use.
        :param uri: The URI to fetch.
        :param params: Query parameters as dict or bytes.
        :param data: Request body contents.

        """
        method = method.lower()

        if data is not None:
            data = json.dumps(data)

        abs_uri = self._make_abs_uri(uri)
        verify_ssl_certs = self.config['verify_ssl_certs']

        headers = {
            'authorization': "Bearer {0}".format(self.access_token),
            'accept': 'application/json;charset=UTF-8',
            'accept-charset': 'UTF-8',
            'content-type': 'application/json',
            'user-agent': "Tictail Python {0}".format(__version__)
        }

        timeout = self.config['timeout']

        try:
            resp = requests.request(method, abs_uri,
                                    params=params,
                                    data=data,
                                    headers=headers,
                                    timeout=timeout,
                                    verify=verify_ssl_certs)

            # `requests` will store an `HTTPError` if one happened.
            resp.raise_for_status()

            content = resp.json() if resp.text else None
            return content, resp.status_code
        except ConnectionError as ce:
            self._handle_connection_error(ce)
        except HTTPError as he:
            self._handle_http_error(he)
        except Exception as e:
            e.response = resp
            self._handle_unexpected_error(e)



"""
tictail.importer
~~~~~~~~~~~~~~~~

Imports various needed dependencies with fallbacks. Provides some help with
installing missing dependencies.

Dependencies:
  * json/simplejson
  * requests

"""

def raise_import_error_with_hint(dep):
    """Raises an import error and gives a helpful message for installing the
    missing dependency.

    :param dep: the missing dependency.

    """
    raise ImportError(("`{0}` is required. "
                       "Install it with `pip install {0}` or "
                       "`easy_install {0}` if you are not using pip. "
                       "For more details, visit http://pip-installer.org.")
                       .format(dep))


# Try to import `requests`.
try:
    import requests
except ImportError:
    raise_import_error_with_hint('requests')


# Try to import `json` and fallback to `simplejson` if not available.
try:
    import json
except ImportError:
    try:
        import simplejson as json
    except ImportError:
        raise_import_error_with_hint('simplejson')



"""
tictail.client
~~~~~~~~~~~~~~

"""

import copy

from .transport import RequestsHttpTransport
from .resource import (Store,
                       Followers,
                       Cards,
                       Customers,
                       Products,
                       Orders,
                       Stores,
                       Theme,
                       Categories,
                       Me)


# API version supported by these bindings.
VERSION = 1

# Default HTTP protocol to use.
DEFAULT_PROTOCOL = 'https'

# Base API URL.
BASE = 'api.tictail.com'

# Whether `requests` should verify SSL certificates.
VERIFY_SSL_CERTS = True

# Default socket timeout.
DEFAULT_TIMEOUT = 20

# Defauly applied configuration.
DEFAULT_CONFIG = {
    'version': VERSION,
    'protocol': DEFAULT_PROTOCOL,
    'base': BASE,
    'verify_ssl_certs': VERIFY_SSL_CERTS,
    'timeout': DEFAULT_TIMEOUT
}


class Client(object):

    def __init__(self, access_token, config=None, transport=None):
        self.access_token = access_token
        self.config = self._make_config(config)

        if transport is None:
            transport = self._make_transport()

        self.transport = transport

    def _make_config(self, config_override):
        config = copy.deepcopy(DEFAULT_CONFIG)
        if config_override:
            config.update(config_override)
        return config

    def _make_transport(self):
        return RequestsHttpTransport(self.access_token, self.config)

    def _make_store_subresource(self, resource_cls, store_id):
        if store_id is None:
            raise ValueError('`store_id` cannot be None')
        parent = "{0}/{1}".format(Store.endpoint, store_id)
        return resource_cls(self.transport, parent=parent)

    def me(self):
        """Alias for getting the store for which the access token you are using
        is valid.

        >>> tt = Tictail('token')
        >>> tt.me()
        Store({...})

        """
        return Me(self.transport).get()

    # ====== Resource factories ======= #

    def stores(self):
        """Returns a `Stores` collection."""
        return Stores(self.transport)

    def followers(self, store=None):
        """Returns a `Followers` collection.

        :param store: A store id.

        """
        return self._make_store_subresource(Followers, store)

    def cards(self, store=None):
        """Returns a `Cards` collection.

        :param store: A store id.

        """
        return self._make_store_subresource(Cards, store)

    def customers(self, store=None):
        """Returns a `Customers` collection.

        :param store: A store id.

        """
        return self._make_store_subresource(Customers, store)

    def products(self, store=None):
        """Returns a `Products` collection.

        :param store: A store id.

        """
        return self._make_store_subresource(Products, store)

    def orders(self, store=None):
        """Returns an `Orders` collection.

        :param store: A store id.

        """
        return self._make_store_subresource(Orders, store)

    def theme(self, store=None):
        """Returns a `Theme` resource.

        :param store: A store id.

        """
        return self._make_store_subresource(Theme, store)

    def categories(self, store=None):
        """Returns a `Categories` collection.

        :param store: A store id.

        """
        return self._make_store_subresource(Categories, store)



#!/usr/bin/env python

from setuptools import setup


with open('requirements.txt') as fd:
    requirements = fd.readlines()


with open('tictail/version.py') as fd:
    version = fd.read().split('=')[1].replace("'", '').strip()


with open('README.md') as fd:
    long_description = fd.read()


setup(
    name='tictail',
    version=version,
    packages=['tictail', 'tictail.resource'],
    license='MIT',
    author='Tictail AB',
    author_email='tech@tictail.com',
    url='https://github.com/tictail/tictail-python',
    download_url='https://github.com/tictail/tictail-python/releases',
    description='Python bindings for the Tictail API',
    keywords=['tictail', 'rest', 'api'],
    install_requires=requirements,
    long_description=long_description,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Topic :: Software Development :: Libraries",
    ]
)